(******************************************************************************
*
* arghandle.ml: parameters handling
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

(* default values for mean field or simulated annealing *)
let k = ref 1.5
let alpha = ref 0.0055
let edgenum = ref 1
let help = ref false
let save = ref false
let syn = ref false
let sa = ref false
let sources = ref []

(* default values for linear logic edge detector *)
let threshold = ref 0.01
let ndirs = ref 16

(* common default values *)
let filename = ref ""

module AnnealingArg = struct

    let specl = [ ("-help", Arg.Set help,
                    "         display this help message");
            ("-k", Arg.Float (fun f -> k := f),
                    "            k value (default = 1.5)");
            ("-alpha", Arg.Float (fun f -> alpha := f),
                    "        alpha value (default = 0.0055)");
            ("-n", Arg.Int (fun f -> edgenum := f),
                    "            minimum number of edge in an edgel (default = 1)");
            ("-s", Arg.Set save,
                    "            save edge map");
            ("-syn", Arg.Set syn, 
                    "          MFA's synchronous version");
            ("-sa", Arg.Set sa, 
                    "           Simulated Annealing algorithm"
            )
            ]

    let usage = "Caml's eye\n"
              ^ "Usage: " ^ Sys.argv.(0) ^ " [options] file\n"
              ^ "  file\t\t ps file containing edges\n\n"
              ^ "Options:"

    let get_par () =
        let anon s = sources := !sources @ [s] in
        let _ = Arg.parse specl anon usage in
        let _ =
            if ( ( List.length !sources ) == 1) then
                let _ = filename := List.hd !sources in
                let len = String.length !filename in
                    if ( len > 3 ) && ( (String.sub !filename ( len - 3 ) 3 ) = ".ps" ) then
                        filename := String.sub !filename 0 (len - 3) 
                    else
                        invalid_arg ( "wrong " ^ !filename ^ " extension." )
                else
                    ( Arg.usage specl usage ; 
                      exit 0 )
        in
        if ( !help ) then (* user needs help *)
                (Arg.usage specl usage; exit 0)
        else ()
end

module LogLinArg = struct

    let specl = [ ("-help", Arg.Set help,
                    "         display this help message");
            ("-ndirs", Arg.Int ( fun n -> ndirs := n ),
                    "        number of directions (only even values) (default = 16)");
            ("-t", Arg.Float ( fun f -> threshold := f ),
                    "            threshold value (default = 0.01)")
            ]

    let usage = "Linear Logic Edge Detector\n"
              ^ "Usage: " ^ Sys.argv.(0) ^ " [options] file\n"
              ^ "  file\t\t (pgm/pbm/ppm) file containing ascii or binary image\n\n"
              ^ "Options:"

    let get_par () =
        let anon s = sources := !sources @ [s] in
        let _ = Arg.parse specl anon usage in
        let _ =
            if ( ( List.length !sources ) == 1) then
                let _ = filename := List.hd !sources in
                let len = String.length !filename in
                    if ( len > 4 ) && 
                        ( 
                            ( (String.sub !filename ( len - 4 ) 4 ) = ".pgm" ) || 
                            ( (String.sub !filename ( len - 4 ) 4 ) = ".pbm" ) || 
                            ( (String.sub !filename ( len - 4 ) 4 ) = ".ppm" ) 
                        ) 
                    then
                        ()
                    else
                        invalid_arg ( "wrong " ^ !filename ^ " extension." )
                else
                    ( Arg.usage specl usage ; 
                      exit 0 )
        in
        if ( !help ) then (* user needs help *)
                (Arg.usage specl usage; exit 0)
        else ()
end

module ParallelArg = struct
    let specl = [ ("-help", Arg.Set help,
                    "         display this help message");
            ("-k", Arg.Float (fun f -> k := f),
                    "            k value (default = 1.5)");
            ("-alpha", Arg.Float (fun f -> alpha := f),
                    "        alpha value (default = 0.0055)");
            ("-n", Arg.Int (fun f -> edgenum := f),
                    "            minimum number of edge in an edgel (default = 1)");
	    	("-p4pg", Arg.Unit (fun () -> ()), "        set procgroup file" );
    		("-p4wd", Arg.Unit (fun () -> ()), "        set working directory\n" )
            ]

    let usage = "Caml's eyes: a parallel version of camleye.\n"
              ^ "Usage: " ^ Sys.argv.(0) ^ " [options] file\n"
              ^ "Options:\n"
              ^ "  file\t\t ps file containing edges"

    let get_par () =
        let anon s = sources := !sources @ [s] in
        let _ = Arg.parse specl anon usage in
        let _ =
            if ( ( List.length !sources ) == 1) then
                let _ = filename := List.hd !sources in
                let len = String.length !filename in
                    if ( len > 3 ) && ( (String.sub !filename ( len - 3 ) 3 ) = ".ps" ) then
                        filename := String.sub !filename 0 (len - 3) 
                    else
                        invalid_arg ( "wrong " ^ !filename ^ " extension." )
                else
                    ( Arg.usage specl usage ; 
                      exit 0 )
        in
        if ( !help ) then (* user needs help *)
                (Arg.usage specl usage; exit 0)
        else ()

end

let annealing_get_parameters ()  =
    let _ = AnnealingArg.get_par () in
    ( !k, !alpha , !edgenum, !filename )

let loglin_get_parameters ()  =
    let _ = LogLinArg.get_par () in
    ( !ndirs, !threshold, !filename )

let parallel_get_parameters ()  =
	let _ = ParallelArg.get_par () in
	( !k, !alpha , !edgenum, !filename )
