(******************************************************************************
*
* edgel.ml : edgels type definition and relative operations.
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

open Edge

module Edgel = 
    struct

        (* Data types and constructors *)

        type point = Point of int * int

        type theta = float

        type magnitude = float

        type gradient =
            Gradient of theta * magnitude

        type edgel =
            Edgel of point * gradient * ( Edge.edge list )
            
        type ndirections = int

        type edgellist = 
            ndirections * edgel list
            
        (* Functions *)
        let lambda
            ( Edgel ( Point ( x1, y1 ), Gradient ( t1, _ ) , _ ) )
            ( Edgel ( Point ( x2, y2 ), Gradient ( t2, _ ) , _ ) )
        =
            let atan a =
                let a = Pervasives.atan a in
                if a < 0.0 then
                    a +. Math.pi *. 0.5
                else
                    a                
            in
            let gamma theta theta1 =
                theta1 -. theta +. ( if theta1 < theta then Math.pi else 0.0 )
            in
            let theta = atan (
                ( float_of_int ( y2 - y1 ) ) /.
                ( float_of_int ( x2 - x1 ) )
            )
            in
            ( gamma theta t1, gamma theta t2 )

        let distance
            ( Edgel ( Point ( x1, y1 ), _, _ ) )
            ( Edgel ( Point ( x2, y2 ), _, _ ) )
        =
            sqrt ( ( float_of_int ( y2 - y1 ) ) ** 2.0 +. ( float_of_int ( x2 - x1 ) ) ** 2.0 )
        
        let string_of_edgel 
            ( Edgel ( Point ( x, y ), Gradient ( t, m ), _ ) )
        =
            Printf.sprintf "Edgel(Point( %d, %d ),Gradient( %f, %f ))" x y t m  

    let edgels_of_edges ( edges : Edge.edgelist ) : edgellist =
        (* ndirs is the number of direction - always even *)
        let ( ndirs, edges ) = edges in
        (* compare function for non vertical edges *)
        let comp1 x2 x1 =
            ( x2 = x1 + 1 )
        in
        (* compare function for vertical edges *)
        let comp1_vertical x2 x1 =
            ( x1 = x2 )
        in
	let comp2 x2 x1 =
	    x2 = x1
	in
	let comp2_vertical x2 x1 =
	    false
	in
        (* gets an edge and an edge list and returns a new edgel and the list of remaining edges *)
        let build_edgel comp1 comp2 ( Edge.Edge ( Edge.Point ( x, y ), Edge.Gradient ( d, i ) ) ) edge_list =            
            (* angle from discrete to continuous *)
            let convert_angle alpha = 
                Math.pi -.
                    mod_float 
                        ( ( float_of_int alpha ) *. Math.pi /. ( float_of_int ( ndirs / 2 ) ) ) 
                        Math.pi
            in 
            (* computes the angular coefficient and it is used as offset *)
            let m =
                if ( d mod ( ndirs / 2 ) = ( ndirs / 4 ) ) then
                    1
                else
                    - ( int_of_float ( ceil ( tan ( convert_angle d ) ) ) )
            in
            (* computes the new coordinate as the midpoint between x1 and x2*)
            let new_point x1 x2 = 
                ( min x1 x2 ) + ( abs( x1 - x2 ) / 2 )
            in
            (* construction of an edgel *)
            let rec _build_edgel 
                    ( Edge.Edge ( Edge.Point ( x1, y1 ), Edge.Gradient ( d1, i1 ) ) )
                    merged_edges
                    sum_i
                    num
                    edge_list
            =
                match edge_list with
                    | [] ->
                        (
                             (
                                 Edgel 
                                    ( 
                                        Point ( ( new_point x x1, new_point y y1 ) ), 
                                        Gradient ( convert_angle d, sum_i /. num ),
                                        merged_edges
                                    )
                             ),
                             []
                        )
                    | ( Edge.Edge 
                        ( 
                            Edge.Point ( x2, y2 ), 
                            Edge.Gradient ( d2, i2 ) 
                        ) 
                      ) :: tail ->
                        let is_same_angle = d2 mod ( ndirs / 2 ) = d mod ( ndirs / 2) in
                        if ( is_same_angle && ( comp1 x2 x1 ) && ( y2 = y1 + m ) ) then
                            (* found a compatible edge *)
                            _build_edgel 
                                ( Edge.Edge ( Edge.Point ( x2, y2 ), Edge.Gradient ( d2, i2 ) ) )
                                ( ( Edge.Edge ( Edge.Point ( x2, y2 ), Edge.Gradient ( d2, i2 ) ) ) :: merged_edges )
                                ( sum_i +. ( float_of_int i2 ) )
                                ( num +. 1.0 )
                                tail
                        else if ( is_same_angle && ( ( ( comp1 x2 x1 ) && ( y2 < y1 + m ) ) || ( comp2 x2 x1 ) ) ) then
                            (* compatible edge not found, but could exist *)
                            let ( el, tail ) =
                                _build_edgel 
                                    ( Edge.Edge (Edge.Point ( x1, y1 ), Edge.Gradient ( d1, i1 ) ) )
                                    merged_edges
                                    sum_i
                                    num
                                    tail
                            in
                            ( el, ( Edge.Edge ( Edge.Point ( x2, y2 ), Edge.Gradient ( d2, i2 ) ) ) :: tail)
                        else
                            (* no more compatible edges can be find *)
                            (
                                (
                                    Edgel 
                                        ( 
                                            Point ( ( new_point x  x1 ),( new_point y  y1 )  ),
                                            Gradient ( convert_angle d,  sum_i /. num ),
                                            merged_edges
                                        )
                                ),
                                edge_list
                            )
            in
            _build_edgel 
                ( Edge.Edge 
                    ( 
                        Edge.Point ( x, y ), 
                        Edge.Gradient ( d, i ) 
                    ) 
                ) 
                [
                    Edge.Edge 
                    ( 
                        Edge.Point ( x, y ), 
                        Edge.Gradient ( d, i ) 
                    ) 
                    
                ]
                ( float_of_int i ) 
                1.0 
                edge_list
        in
        (* convert an Edge list into an Edgel list *)
        let rec _to_edgels edges =
            match edges with
                | []  ->
                    []
                | ( Edge.Edge ( Edge.Point ( x, y ), Edge.Gradient ( d, i ) ) ) :: tail  ->
                    let ( edgel, tail ) =
                        if ( d mod ( ndirs / 2 ) = ( ndirs / 4 ) ) then
                            build_edgel 
                                comp1_vertical 
				comp2_vertical
                                ( Edge.Edge 
                                    ( 
                                        Edge.Point ( x, y ), 
                                        Edge.Gradient ( d, i ) 
                                    ) 
                                ) 
                                tail
                        else
                            build_edgel 
                                comp1
				comp2 
                                ( 
                                    Edge.Edge 
                                        ( 
                                            Edge.Point ( x, y ), 
                                            Edge.Gradient ( d, i ) 
                                        ) 
                                ) 
                                tail
                    in
                    edgel :: ( _to_edgels tail )
        in
        (* sorts edge list *)
        let edges = Edge.sort_edges ( ndirs, edges ) in
        (ndirs, (_to_edgels edges) )

    let edges_of_edgels ( ( ndirs, edgels ) : edgellist ) : Edge.edgelist =
        let rec _edges_of_edgels edgels =
            match edgels with
                | [] ->
                    []
                | ( Edgel ( _, _, merged_edges ) ) :: tail -> 
                    merged_edges @ ( _edges_of_edgels tail )
        in
        ( ndirs, _edges_of_edgels edgels )

    end
