(******************************************************************************
*
* image.ml : simple grayscale image management
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

open Error

module Image = struct 

    (* Data types and constructors *)

    type width = int

    type height = int

    (* number of gray tones *)
    type depth = int

    type image =
        Image of width * height * depth * int array array

    (* Functions *)

    let get ( Image ( width, height, _, imagemap ) ) x y =
        ( sassert ( x >= 0 && x < width ) "get: x coordinate out of bounds" ;
          sassert ( y >= 0 && y < height ) "get: y coordinate out of bounds" ;
          imagemap.(y).(x) )

    let set image x y value =
        match image with
        | ( Image ( _, _, depth, imagemap ) ) ->
            (
                ( sassert ( value >= 0 && value <= depth ) "pixel value exceeds depth" ) ;
                ( imagemap.(y).(x) <- value ; image )
            )

    let get_dimensions ( Image ( width, height, _, _) ) =
        ( width, height )

    let get_depth ( Image ( _, _, depth, _) ) =
        depth

    let init width height depth f =
        let safe_f x y =
            let value = f x y in
            ( sassert ( value <= depth ) "pixel value exceeds depth" ; value )
        in
        let imagemap = 
            Array.init height ( fun y -> 
                Array.init width ( fun x -> safe_f x y 
                )
            )
        in
        Image ( width, height, depth, imagemap )

    let init_bg width height depth bgcolor =
        init width height depth ( fun x y -> bgcolor )

    let grayscale r g b =
        (* CIE XYZitu (D65) *)
        (* ( 222 * r + 707 * g + 71 * b ) / 1000 *)
        (* YCbCr ITU *)
        ( 2125 * r + 7154 * g + 721 * b ) / 10000

    let transform image f =
        match image with 
            | Image ( width, height, depth, imagemap ) ->
                init width height depth ( f image )
    
    let array_of_image ( Image ( _, _, _, imagemap ) ) =
        imagemap

    let image_of_array a =
        let width = Array.length ( a.(0) ) in
        let height = Array.length a in
        let depth =
            Array.fold_left ( fun b row -> max b ( Array.fold_left max 0 row ) ) 0 a
        in
        Image ( width, height, depth, a )

    let d2x image x y =
        get image ( x - 1 ) y -
        2 * ( get image x y ) +
        get image ( x + 1 ) y

    let d2y image x y =
        get image x ( y - 1 ) -
        2 * ( get image x y ) +
        get image x ( y + 1 )

    (* apply mask to imagemap *)
    let apply_filter
        ( Image ( mw, mh, _, mask ) )
        ( Image ( width, height, depth, imagemap ) )
    =
    let center x = x / 2 in
    let ( xc, yc ) = ( center mw, center mh ) in
    (* apply mask to point x y of imagemap *)
    let convolve_point x y =
        let rec convolve_point_r mx my acc =
            if my = mh then
                acc
            else if mx = mw then
                convolve_point_r 0 ( my + 1 ) acc
            else
                convolve_point_r ( mx + 1 ) my
                (
                    acc + (
                        if 
                            x + mx - xc >= 0 && x + mx - xc < width &&
                            y + my - yc >= 0 && y + my - yc < height
                        then
                            imagemap.( y + my - yc ).( x + mx - xc )
                        else if x >= width or y >= height then
                            0
                        else
                            imagemap.(y).(x) 
                    ) * mask.(my).(mx)
                )
        in
        ( convolve_point_r 0 0 0 )
    in
    (* convolve: *)
    let maxval = ref 0 in
    let imagemap = 
        Array.init height ( fun y -> 
            Array.init width (
                fun x -> 
                    let _ = Printf.printf "%dx%d \n%!" x y in
                    let v = convolve_point x y in
                        ( maxval := max !maxval v ; v ) )
        )
    in
    (* normalize: *)
    let _ =
        Array.iter (
            fun row -> 
                Array.iteri ( 
                    fun i a -> 
                        row.(i) <- 
                            if a > 0
                                then a * depth / !maxval 
                                else 0 )
                row )
            imagemap
    in
    Image ( width, height, depth, imagemap )

    (* convolution operator *)
    let convolve f g =
        match ( f, g ) with
            | ( Image ( w1, h1, _, _ ), Image ( w2, h2, _, _ ) ) 
              when w1 <= w2 && h1 <= h2 ->
                apply_filter f g
            | ( Image ( w1, h1, _, _ ), Image ( w2, h2, _, _ ) ) 
              when w2 <= w1 && h2 <= h1 ->
                apply_filter g f
            | ( Image ( w1, h1, _, _ ), Image ( w2, h2, _, _ ) ) ->
                (* center images: *)
                apply_filter f g
    
    (* Some simple masks *)
    module Masks = struct
    
        let edge_detect = image_of_array
        [|
            [| 1;   4; 1 |];
            [| 4; -20; 4 |];
            [| 1;   4; 1 |]
        |]
        
        let find_hor_edge = image_of_array
        [|
            [| -1; -1; -1 |];
            [|  0;  0;  0 |];
            [|  1;  1;  1 |];
        |]

        let soften_h = image_of_array
        [|
            [| 1 ; 4 ; 6 ; 4 ; 1 |]
        |]
        
        let soften_v = image_of_array
        [|
            [| 1 |];
            [| 4 |];
            [| 6 |];
            [| 4 |];
            [| 1 |]            
        |]

    end

end