(******************************************************************************
*
* interactionCoefficients.ml : computes the interaction coefficients between
*                              edgels.
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
*******************************************************************************)

open Edgel
open Error
open SymmetricMatrix

module InteractionCoefficients = struct

let pi2 = Math.pi ** 2.0

(* interaction between two edgels *)
let interaction k sigma_d2 g_max2 d e1 e2 =
    let ( lambda1, lambda2 ) = Edgel.lambda e1 e2 in
    let delta2 = ( abs_float ( lambda1 +. lambda2 -. Math.pi ) ) ** 2.0 in
    let d2 = d ** 2.0 in
    match ( e1, e2 ) with
        | (
            Edgel.Edgel ( p1, Edgel.Gradient ( _, g1 ), _ ),
            Edgel.Edgel ( p2, Edgel.Gradient ( _, g2 ), _ ) 
          ) ->
            if p1 = p2 then
                0.0
            else
                (* cocircularity *)
                ( ( 1.0 -. delta2 /. pi2 ) *. exp ( -. delta2 /. k  ) ) *.
                (* smoothness *)
                (
                    ( 1.0 -. lambda1 *. ( Math.pi -. lambda1 ) /. pi2 ) *.
                    ( 1.0 -. lambda2 *. ( Math.pi -. lambda2 ) /. pi2 )
                ) *.
                (* proximity *)
                ( exp ( -. d2 /. sigma_d2 ) ) *.
                (* contrast *)
                ( g1 *. g2 /. g_max2 )

(* compute preliminar interaction matrix *)
let interaction_matrix interaction_fun matrix n edgels =
    (* HELPER *)
    let fill_matrix matrix n edgels =
        let rec fill_matrix_r matrix i j row_edgels col_edgels =
            if i = n then
                matrix
            else if j = n then
                let row_edgels = List.tl row_edgels in
                fill_matrix_r matrix ( i + 1 ) ( i + 1 ) row_edgels row_edgels
            else
                let matrix =
                    if i = j then
                        SymmetricMatrix.set matrix i j 0.0
                    else
                        let c_ij =
                                interaction_fun
                                    (* this is the distance *)
                                    ( SymmetricMatrix.get matrix i j )
                                    ( List.hd row_edgels ) ( List.hd col_edgels )
                        in
                        let _ = sassert ( c_ij >= 0.0 ) "an interaction coefficient is negative" in
                        SymmetricMatrix.set matrix i j c_ij
                in
                fill_matrix_r
                    matrix i ( j + 1 )
                    row_edgels ( List.tl col_edgels )
        in
        fill_matrix_r matrix 0 0 edgels edgels
    in
    (* calculates interaction matrix *)
    let m = fill_matrix matrix n edgels in
    m

(* creates distance matrix *)
let distance_matrix sigma_d_coeff matrix n edgels =
    let rec distance_matrix_r
        matrix i j row_edgels col_edgels sum_x sum_x2
    =
        if i = n then
            let n_distances = float_of_int ( n * ( n - 1 ) / 2 ) in
            (
                sigma_d_coeff *.
                    (* standard deviation *)
                    sqrt
                    (
                        (
                            sum_x2
                            -.
                            ( ( sum_x ** 2.0 ) /. n_distances )
                        )
                        /.
                        n_distances
                    ),
                matrix
            )
        else if j = n then
            let row_edgels = List.tl row_edgels in
            distance_matrix_r
                matrix
                ( i + 1 ) ( i + 1 ) 
                row_edgels row_edgels
                sum_x sum_x2
        else
            let distance = Edgel.distance ( List.hd row_edgels ) ( List.hd col_edgels ) in
            let matrix = SymmetricMatrix.set matrix i j distance in
            distance_matrix_r
                matrix
                i ( j + 1 )
                row_edgels ( List.tl col_edgels )
                ( sum_x +. distance ) ( sum_x2 +. ( distance ** 2.0 ) )
    in
    distance_matrix_r matrix 0 0 edgels edgels 0.0 0.0

end

open InteractionCoefficients

(*
  get_energy_function_params, returns:
  - the interaction coefficients matrix J;
  - the external fields delta;
*)
let get_energy_function_params ~k ?(sigma_d_coeff = 1.0) ~alpha edgels =
    (* HELPER *)
    let build_delta c n =
        let nalpha = ( float_of_int n ) *. alpha in
        let rec build_delta_r building_delta i j c =
            if i = n then
                building_delta
            else if j = n then
                let _ =
                    building_delta.(i) <-                                      (* SIDEFX *)
                        ( building_delta.(i) -. nalpha ) /. 2.0
                in
                build_delta_r building_delta ( i + 1 ) 0 c
            else
                let _ =
                    building_delta.(i) <-                                      (* SIDEFX *)
                            building_delta.(i) +.
                            ( SymmetricMatrix.get c i j )
                in
                build_delta_r building_delta i ( j + 1 ) c
        in
        build_delta_r ( Array.init n ( fun x -> 0.0 ) ) 0 0 c
    in
    let n = List.length edgels in
    (* find max gradient magnitude *)
    let g_max =
        List.fold_left
            (
                fun g_max ( Edgel.Edgel ( _, Edgel.Gradient ( _, g ), _ ) ) ->
                    max g_max g               
            )
            0.0
            edgels
    in
    (* precomputes g_max^2 *)
    let g_max2 = g_max ** 2.0 in
    let _ = sassert ( g_max2 > 0.0 ) "g_max (max intensity) is negative" in
    (* allocates matrix *)
    let matrix = ( SymmetricMatrix.make_matrix n 0.0 ) in
    (* precomputes distance matrix and gets standard deviation *)
    let ( sigma_d, matrix ) = distance_matrix sigma_d_coeff matrix n edgels in
    let _ = sassert ( sigma_d > 0.0 ) "sigma_d (distance variance) is negative" in
    (* precomputes sigma_d^2 *)
    let sigma_d2 = 2.0 *. ( sigma_d ** 2.0 ) in
    (* partially apply interaction function *)
    let interaction_fun = interaction k sigma_d2 g_max2 in
    (* preliminar interaction coefficients matrix *)
    let c = interaction_matrix interaction_fun matrix n edgels in
    (* external fields *)
    let delta = build_delta c n in
    (* true interaction coefficients matrix *)
    let _ =
            SymmetricMatrix.iteri
            (
                fun i row ->
                        SymmetricMatrix.iteri
                        ( fun j c_ij -> if i != j then row.(j) <- ( c_ij -. alpha ) /. 2.0 else () )
                        row
            )
            c
    in
    ( c, delta )

(* create matrix of coefficients *)
let coefficients_matrix ~k ?(sigma_d_coeff = 1.0) edgels =
    let n = List.length edgels in
    (* find max gradient magnitude *)
    let g_max =
        List.fold_left
            (
                fun g_max ( Edgel.Edgel ( _, Edgel.Gradient ( _, g ), _ ) ) ->
                    max g_max g               
            )
            0.0
            edgels
    in
    (* precomputes g_max^2 *)
    let g_max2 = g_max ** 2.0 in
    let _ = sassert ( g_max2 > 0.0 ) "g_max (max intensity) is negative" in
    (* allocates matrix *)
    let matrix = ( SymmetricMatrix.make_matrix n 0.0 ) in
    (* precomputes distance matrix and gets standard deviation *)
    let ( sigma_d, matrix ) = distance_matrix sigma_d_coeff matrix n edgels in
    let _ = sassert ( sigma_d > 0.0 ) "sigma_d (distance variance) is negative" in
    (* precomputes sigma_d^2 *)
    let sigma_d2 = 2.0 *. ( sigma_d ** 2.0 ) in
    (* partially apply interaction function *)
    let interaction_fun = interaction k sigma_d2 g_max2 in
    (* preliminar interaction coefficients matrix *)
    let c = interaction_matrix interaction_fun matrix n edgels in
    c
