(******************************************************************************
*
* logicalLinear.ml : detects edges with logical linear operators.
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

open Edge
open Error
open Image
open ProgressNotifier
open Utils

module LogicalLinear = struct

(* TYPE DEFINITIONS *)
(* mask, maxval, xc, yc *)
type fmask = ( float array array * float * int * int )

(* mask, divisor, xc, yc *)
type imask = ( int array array * int * int * int )

(* operator type definition *)
type 'a operator =
    | LlandN of 'a operator list
    | LlorN of 'a operator list
    | Phi of 'a

(* a linear operator has \phi(x): R -> R *)
type linear_operator = ( float -> float ) operator

(* a convolution operator has \phi(x, y): S -> S, S \in N^2 *)
type convolution_operator = imask operator

(* FUNCTIONS *)

(* sigmoid *)
let partition ro x =
    let x = ( x +. 1.0 /. ( 2.0 *. ro ) ) *. ro in
    if x < 0.0 then 0.0
    else if x > 1.0 then 1.0
    else x

(* linear logical and *)
let lland ro xs =
    let sigmas = List.map ( partition ro ) xs in
    let sigmas_prod = List.fold_left ( *. ) 1.0 sigmas in
    let single_sum res x s = res +. x *. ( sigmas_prod +. ( 1.0 -. s ) ) in
    List.fold_left2 single_sum 0.0 xs sigmas

let abs_lland xs =
    let abs_lland x y =
        if x > 0 || y <= 0 then y else 0 +
        if x <= 0 || y > 0 then x else 0
    in
    List.fold_left abs_lland ( List.hd xs ) ( List.tl xs )

(* linear logical or *)
let llor ro xs =
    let sigmas = List.map ( fun x -> partition ro ( -. x ) ) xs in
    let sigmas_prod = List.fold_left ( *. ) 1.0 sigmas in
    let single_sum res x s = res +. x *. ( sigmas_prod +. ( 1.0 -. s ) ) in
    List.fold_left2 single_sum 0.0 xs sigmas

let abs_llor xs =
    let abs_llor x y =
        if x > 0 || y <= 0 then x else 0 +
        if x <= 0 || y > 0 then y else 0
    in
    List.fold_left abs_llor ( List.hd xs ) ( List.tl xs )

(* combines two linear operator basis (xop \times yop) to have a convolution operator *)
let cartesian n angle yop op =
    (* creates an operator for ( f \times op ) with masks of size n x n *)
    let rec cartesian_f n angle f = function
        | LlandN ( ops ) ->
            LlandN ( List.map ( cartesian_f n angle f ) ops )
        | LlorN ( ops ) ->
            LlorN ( List.map ( cartesian_f n angle f ) ops )
        | Phi g ->
            let max = ref 0.0 in
            let ( xc, yc ) = ( n / 2, n / 2 ) in
            let mask =
                Array.init n ( fun y ->
                    Array.init n ( fun x ->
                        let ( c, s ) = ( cos angle, sin angle ) in
                        let ( x, y ) = ( float_of_int ( x - xc ), float_of_int ( y - yc ) ) in
                        let ( x', y' ) = ( c *. x +. s *. y, -. s *. x +. c *. y ) in
                        let v = ( f x' ) *. ( g y' ) in
                        ( max := if abs_float v > abs_float !max then v else !max ; v )
                    )
                )
            in
            Phi ( mask, !max, xc, yc )
    in
    (* combines two linear operator basis (xop \times yop) to have a convolution operator *)
    let rec cartesian_r n angle yop = function
        | LlandN ( ops ) ->
            LlandN ( List.map ( cartesian_r n angle yop ) ops )
        | LlorN ( ops ) ->
            LlorN ( List.map ( cartesian_r n angle yop ) ops )
        | Phi f ->
            cartesian_f n angle f yop
    in
    cartesian_r n angle yop op

(* normalizes operator to have the maximum positive response = 1.0 and shrinks masks *)
let normalize operator =
    (* gets out all masks of an operator *)
    let get_masks operator = 
        let merge_compatibles ( l1, w1, h1 ) ( l2, w2, h2 ) =
            ( Error.sassert ( w1 = w2 && h1 = h2 ) "logicalLinear.normalize: kernels are incompatible" ;
              ( l1 @ l2, w1, h1 ) )
        in
        let rec get_masks_r = function
            | LlandN ( ops ) ->
                let masks = List.map get_masks_r ops in
                List.fold_left merge_compatibles ( List.hd masks ) ( List.tl masks )
            | LlorN ( ops ) ->
                let masks = List.map get_masks_r ops in
                List.fold_left merge_compatibles ( List.hd masks ) ( List.tl masks )
            | Phi ( mask, _, _, _ ) ->
                ( [ mask ], Array.length mask, Array.length mask.(0) )
        in
        get_masks_r operator
    in
    (* finds maximum positive response in a point for a list of masks *)
    let rec max_positive_response_r w h x y acc masks =
        let rec point_max x y acc = function
            | [] ->
                acc
            | head :: tail ->
                point_max x y ( acc +. head.(y).(x) ) tail
        in
        if y = h then
            acc
        else if x = w then
            max_positive_response_r w h 0 ( y + 1 ) acc masks
        else
            let v = point_max x y 0.0 masks in
            max_positive_response_r w h ( x + 1 ) y ( acc +. ( if v > 0.0 then v else 0.0 ) ) masks
    in
    let rec normalize_r scale = function
        | LlandN ( ops ) ->
            LlandN ( List.map ( normalize_r scale ) ops )
        | LlorN ( ops ) ->
            LlorN ( List.map ( normalize_r scale ) ops )
        | Phi ( mask, maxval, xc, yc ) ->
            (* thresholds and shrinks matrixes *)
            let cols_zeros = Array.init ( Array.length mask ) ( fun x -> true ) in
            let rows_zeros = Array.init ( Array.length mask.(0) ) ( fun x -> true ) in
            (* thresholds over maxval / 127: *)
            let fmask =
                Array.mapi ( fun j row ->
                    Array.mapi ( fun i x -> 
                        let v = scale *. x in
                        if abs_float v > ( abs_float maxval ) /. 127.0
                            then ( rows_zeros.(i) <- false ;  
                                   cols_zeros.(j) <- false ; 
                                   v )
                            else 0.0
                    ) row
                ) mask
            in
            (* finds borders of only zeros: *)
            let count_consecutive_zeros ( b, i ) x = 
                ( b && x, i + ( if b && x then 1 else 0 ) )
            in
            let ( zleft, zright, ztop, zbottom ) =
                (
                    snd ( Array.fold_left count_consecutive_zeros ( true, 0 ) rows_zeros ) ,
                    snd ( Array.fold_right ( fun x y -> count_consecutive_zeros y x ) rows_zeros ( true, 0 ) ) ,
                    snd ( Array.fold_left count_consecutive_zeros ( true, 0 ) cols_zeros ) ,
                    snd ( Array.fold_right ( fun x y -> count_consecutive_zeros y x ) cols_zeros ( true, 0 ) )
                )
            in
            let ( zright, zbottom ) =
                (
                    ( Array.length rows_zeros ) - 1 - zright,
                    ( Array.length cols_zeros ) - 1 - zbottom
                )
            in
            (* shrinks mask: *)
            let shrinked_fmask =
                Array.init ( zbottom - ztop + 1 ) (
                    fun j -> 
                        Array.init ( zright - zleft + 1 ) (
                            fun i ->
                                fmask.(j + ztop).(i + zleft)
                        )
                )
            in
            Phi ( shrinked_fmask, scale *. maxval, xc - zleft, yc - ztop )
    in
    let ( masks, w, h ) = get_masks operator in
    let max_positive_response = max_positive_response_r w h 0 0 0.0 masks in
    normalize_r ( 1.0 /. max_positive_response ) operator

(* eval an integer convolution operator at a given point in the image *)
let evali_operator ?(ro = 32.0) operator ( Image.Image ( w, h, depth, imagemap ) ) x y =
    let between x min max = if x < min || x > max then false else true in
    let convolve_point ( imask, divisor, xc, yc ) =
        let ( start_mx, start_my ) = ( max 0 ( xc - x ), max 0 ( yc - y ) ) in
        let ( mw, mh ) = ( min ( Array.length imask.(0) ) ( w + xc - x ), min ( Array.length imask ) ( h + yc - y ) ) in
        let ( start_x, start_y ) = ( x + start_mx - xc, y + start_my - yc ) in
        let rec convolve_point_r my y acc =
            let rec convolve_point_r2 mx x acc =
                if mx = mw then
                    acc
                else
                    convolve_point_r2 ( mx + 1 ) ( x + 1 ) ( acc + imagemap.(y).(x) * imask.(my).(mx) )
            in
            if my = mh then
                ( float_of_int acc ) /. ( float_of_int divisor ) /. ( float_of_int depth )
            else
                convolve_point_r ( my + 1 ) ( y + 1 ) ( convolve_point_r2 start_mx start_x acc )
        in
        convolve_point_r start_my start_y 0
    in
    let rec evali_operator_r = function
        | LlandN ( ops ) ->
            let ( vals1, vals2 ) = List.split ( List.map evali_operator_r ops ) in
            ( lland ro vals1, llor ro vals2 ) 
        | LlorN ( ops ) ->
            let ( vals1, vals2 ) = List.split ( List.map evali_operator_r ops ) in
            ( llor ro vals1, lland ro vals2 ) 
        | Phi f ->
            let v = convolve_point f in
            ( v, v )
    in
    let ( v1, v2 ) = evali_operator_r operator in
    ( v1, -. v2 )

(* converts a floating point mask to an integer mask *)
let imask_of_fmask ?(imax = 127) ( fmask, maxval, xc, yc ) =
    let scale = ( float_of_int imax ) /. ( abs_float maxval ) in
    let imask =
        Array.map ( fun row ->
            Array.map ( fun v -> Math.round ( v *. scale ) ) row
        ) fmask
    in
    ( imask, int_of_float scale, xc, yc )

(* writes the operator to the out_channel *)
let rec write_operator operator out_channel =
    let write_imask ( imask, divisor, xc, yc ) out_channel =
        ( Printf.fprintf out_channel "kern %d %d %d %d %d" ( Array.length imask.(0) ) ( Array.length imask ) xc yc divisor ;
          Array.iter ( fun row ->
              ( Printf.fprintf out_channel "\n" ;
                Array.iter ( fun v -> Printf.fprintf out_channel " %4d" v ) row )
          ) imask
        )
    in
    match operator with
        | LlandN ( ops ) ->
            ( List.iter ( fun op -> write_operator op out_channel ) ops ;
              Printf.fprintf out_channel "and %d\n" ( List.length ops ) )
        | LlorN ( ops ) ->
            ( List.iter ( fun op -> write_operator op out_channel ) ops ;
              Printf.fprintf out_channel "or %d\n" ( List.length ops ) )
        | Phi fmask ->
            ( write_imask ( imask_of_fmask fmask ) out_channel ;
              Printf.fprintf out_channel "\nconv\n" )

(* from a floating point convolution operator to an integer convolution operator *)
let rec ioperator_of_foperator ?(imax = 127) = function
    | LlandN ( ops ) ->
        LlandN ( List.map ( ioperator_of_foperator ~imax:imax ) ops )
    | LlorN ( ops ) ->
        LlorN ( List.map ( ioperator_of_foperator ~imax:imax ) ops )
    | Phi fmask ->
        Phi ( imask_of_fmask ~imax:imax fmask )

(* write out kernels *)
let write_kernels ?(filename = "kernel-dir") size tangential_op edge_op ndirs =
    let rec write_kernels_r i =
        if i = ndirs then
          ()
        else
            let angle = ( float_of_int i ) *. Math.pi /. ( float_of_int ndirs ) in
            let op = normalize ( cartesian size angle edge_op tangential_op ) in
            ( Utils.temp_resource
                ( write_operator op ) 
                ( open_out_bin ( filename ^ string_of_int i ^ ".kern" ) )
                close_out ;
              write_kernels_r ( i + 1 ) )
    in
    write_kernels_r 0

(* find edges in an image *)
let find_edges ?(threshold = 0.01) size tangential_op edge_op image ndirs : Edge.edgelist =
    let _ = sassert ( ( ndirs / 2 ) * 2 = ndirs ) "find_edges: needs an even number of directions" in
    let ndirs = ndirs / 2 in
    let ( width, height, depth, imagemap ) = ( fun ( Image.Image ( w, h, d, i ) ) -> ( w, h, d, i ) ) image in
    let progress = ProgressNotifier.make_loglin "Logical Linear Detection..." ( 1, ( height - 2 ) * ndirs ) in
    (* normalizes ro on max image contrast *)
    let ro = 
        let ( min, max ) =
            Array.fold_left ( fun b row ->
                Array.fold_left (
                    fun ( min, max ) x ->
                        ( ( if x < min then x else min ) , ( if x > max then x else max ) )
                ) b row
            ) ( depth, 0 ) imagemap
        in
        32.0 /. ( float_of_int ( max - min ) )
    in
    let rec find_edges_r i acc =
        let scan_image op_eval acc =
            (* adds an edge to the accumulating edge list: *)
            let add_edge x y v dir edge_list = 
                if v > threshold then
                    Edge.Edge (
                        Edge.Point ( x, y ),
                        Edge.Gradient ( dir, int_of_float ( v *. 255.0 ) )
                    ) :: edge_list
                else
                    edge_list
            in
            let rec scan_image_r x y acc =
                if y = height - 1 then
                    acc
                else if x = width - 1 then
                    let _ = progress ( y + i * ( height - 2 ) ) in
                    (* scans rows: let _ = Printf.printf "\r  row %d...%!" y in *)
                    scan_image_r 1 ( y + 1 ) acc
                else
                    (* evals operator for both verses of a direction: *)
                    let ( v1, v2 ) = op_eval x y in
                    scan_image_r ( x + 1 ) y ( add_edge x y v2 ( i + ndirs ) ( add_edge x y v1 i acc ) )
            in
            scan_image_r 1 1 acc
        in
        if i = ndirs then
            acc
        else
            let angle = ( float_of_int i ) *. Math.pi /. ( float_of_int ndirs ) in
            let op = ioperator_of_foperator ( normalize ( cartesian size angle edge_op tangential_op ) ) in
            (* scans directions: let _ = Printf.printf "\ndirection %d\n%!" i in *)
            find_edges_r ( i + 1 ) ( scan_image ( evali_operator op image ) acc )
    in
    (
        ndirs * 2,
        find_edges_r 0 []
    )

end