(******************************************************************************
*
* meanFieldAnnealing.ml: Mean Field Annealing algorithm for the Herault-Horaud
*                        combinatorial figure-ground discrimination.
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

open Error
open SymmetricMatrix
open ProgressNotifier
open Utils

module MeanFieldAnnealing =
    struct
        type state = float array
        type spin = Up | Down

        let _ = Random.init ( int_of_float ( Unix.time () ) )

        (* sum_col : sums over the columns of the matrix j_matrix*)
        let rec sum_col j_matrix (mu : state) i =
            let rec _sum_col j acc =
            if j = SymmetricMatrix.length j_matrix then
                acc
            else
                _sum_col ( j + 1 ) ( ( ( SymmetricMatrix.get j_matrix i j ) *. mu.(j) ) +. acc )
            in
            let _ = assert ( ( SymmetricMatrix.length j_matrix ) > 0 ) in
            _sum_col 0 0.0
		
        (* mu_new : computes the new value for the i-th entry of state vector mu *)
        let mu_new temp delta_i j_matrix (mu : state) i =
            let sum = sum_col j_matrix mu i
            in
            tanh ( ( sum +. delta_i ) /. temp )

        (* set_mu : sets the i-th entry of the vector mu_old*)
        let set_mu ( mu_old : state ) mu_i i =
            let _ = mu_old.(i) <- mu_i   (* SIDEFX *)
            in
            mu_old

        (* update_asyn : updates the state vector mu in an asyncronous mode*)
        let update_asyn temp delta j_matrix (mu_old : state)  =
            let n = Random.int ( Array.length mu_old ) in
            Array.init 
                ( Array.length mu_old ) 
                ( fun i ->
                    if i = n then
                        mu_new temp delta.(i) j_matrix mu_old i
                    else
                        mu_old.(i)
                )

        (* update_syn : updates the state vector mu in an syncronous mode *)
        let update_syn temp delta j_matrix (mu_old : state) : state =
            Array.init 
                ( Array.length mu_old ) 
                ( fun i ->
                        mu_new temp delta.(i) j_matrix mu_old i
                )

        (* sum_mu : sums over the elements of mu*)
        let sum_mu mu =
            Array.fold_left (fun building_sum el ->
                building_sum +. (abs_float el)
            ) 0.0 mu

        (* normalize_mu : creates a spin vector from a state vector *)
        let normalize_mu (mu : state) =
            (* HELPER *)
            let convert el =
                match (el > 0.0) with
                    | true  ->
                        Up
                    | false ->
                        Down
            in
            let rec _normalize_mu i =
                if ( i = ( Array.length mu ) - 1 ) then
                    ( convert mu.(i) ) :: []
                else
                    ( convert mu.(i) ) :: ( _normalize_mu ( i + 1 ) )
            in
            (* mu vector normalization *)
            _normalize_mu 0

    end

open MeanFieldAnnealing

    let mean_field_annealing temp dec_temp step update_fun ( mu : state ) delta j_matrix  =
        let n = Array.length mu in
        let threshold = 0.99 *.  ( float_of_int ( step + 1 ) ) *. (float_of_int n ) in
        let progress_notifier = ProgressNotifier.make "Annealing..." ( ( ( step + 1 ) * int_of_float ( sum_mu mu ) ), ( int_of_float threshold ) ) in
        let rec _mean_field_annealing temp iter mu_iter sum_iter =
            let sum_curr_mu = sum_mu mu_iter in
            let sum_curr_iter = sum_iter +. sum_curr_mu in
            let temp' = temp *. dec_temp in
            let _ = Printf.printf "\riteration: %d %!" iter in
            if ( ( iter mod ( step + 1 ) ) = 0 ) then
                if not ( sum_curr_iter < threshold )
                    then
                        let _ = Printf.printf "\n%!" in
                        ( iter, ( normalize_mu mu_iter ) ) (* base of induction *)
                    else
                        let _ = progress_notifier iter ( int_of_float sum_curr_iter ) in
                        let mu_iter = update_fun temp delta j_matrix mu_iter
                        in
                        _mean_field_annealing temp' ( iter + 1 ) mu_iter 0.0 (* recursion step I *)
            else
                let mu_iter = update_fun temp delta j_matrix mu_iter
                in
                _mean_field_annealing temp' ( iter + 1 ) mu_iter sum_curr_iter (* recursion step II *)
        in
        let _ =
            sassert
                ( ( dec_temp > 0.0 ) && ( dec_temp < 1.0 ) )
                "decrement of temperature must be expressed with a float in the neighbourhood [0..1]"
        in
        _mean_field_annealing temp 1 mu 0.0

    let mean_field_annealing_syn delta j_matrix =
        let mu = Array.init ( Array.length delta) ( fun i -> ( Random.float 2.0 *. 10.0 ** ( -. 5.0 ) ) -. 10.0 ** ( -. 5.0 )) in
        mean_field_annealing 100.0 0.9995 5 update_syn mu delta j_matrix


    let mean_field_annealing_asyn delta j_matrix =
        let mu = Array.init ( Array.length delta ) ( fun i -> ( Random.float 2.0 *. 10.0 ** ( -. 5.0 ) ) -. 10.0 ** ( -. 5.0 )) in
        mean_field_annealing 100.0 0.9995 5 update_asyn mu delta j_matrix

    (* get_figure : selects figure edgels  *)
    let rec get_figure spin_vector edgels =
        match ( spin_vector, edgels ) with
            | ( Up :: spin_vector, edgel :: edgels ) ->
                edgel :: ( get_figure spin_vector edgels )
            | ( Down :: spin_vector, edgel :: edgels ) ->
                ( get_figure spin_vector edgels )
            | ( [], [] ) ->
                []
            | _ ->
                let _ = sassert false "get_figure: spin_vector and edgels sizes differ" in []

