(******************************************************************************
*
* netpbm.ml : netpbm file formats loader ( pbm / pgm / ppm )
*             - ppm is converted to grayscale
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

open Error
open Image

module Netpbm = struct 

    exception Invalid_format of string

    let load in_channel =
        (* Stream utilities *)
        let stream = Stream.of_channel in_channel in
        let junk () =
            Stream.junk stream
        in
        let eof () = 
            Stream.peek stream = None
        in
        let get_char () =
            try
                Stream.next stream
            with
                Stream.Failure ->
                    raise ( Invalid_format "unexpected end of file" )
        in
        let peek_char () =
            match Stream.peek stream with
                | Some c ->
                    c
                | None ->
                    raise ( Invalid_format "unexpected end of file" )
        in
        let peek s =
            match Stream.peek stream with
                | Some c ->
                    String.contains s c
                | None ->
                    false
        in
        (* Reading utilities *)
        let nextbit = ref 8 in
        (* skips a whole line *)
        let rec skip_line () =
            ( junk () ; if not ( peek "\n" ) && not ( eof () ) then skip_line () )
        in
        (* skips all spacers, including comments *)
        let rec skip_spaces () =
            if peek " \t\r\n" then
                ( junk () ; skip_spaces () )
            else if peek "#" then
                ( skip_line () ; skip_spaces () )
            else
                ()
        in
        (* reads a positive decimal integer *)
        let read_dec () =
            let dec_of_char c =
                ( Char.code c ) - ( Char.code '0' )
            in
            let rec read_dec_r acc =
                if peek "0123456789" then
                    read_dec_r ( acc * 10 + dec_of_char ( get_char () ) )
                else
                    acc
            in
            ( skip_spaces () ; read_dec_r 0 )
        in
        (* reads a byte *)
        let read_byte () =
            int_of_char ( get_char () )
        in
        (* reads a packed bit *)
        let read_bit () =
            ( if !nextbit = 0 then ( nextbit := 8 ; junk () ) ;
              nextbit := !nextbit - 1 ;
              1 - ( ( ( int_of_char ( peek_char () ) ) lsr !nextbit ) ) land 1 )
        in
        (* reads rgb color with grayscale conversion *)
        let read_rgb reader () =
            let r = reader () in
            let g = reader () in
            let b = reader () in
            Image.grayscale r g b
        in
        (* returns ( width, height, depth, reader ) *)
        let parse_header =
            let return depth reader =
                let width = read_dec () in
                let height = read_dec () in
                let depth = if depth = 0 then read_dec () else depth in
                if depth < 65536 then
                    ( width, height, depth, reader )
                else
                    raise ( Invalid_format "depth must be less than 65536" )
            in
            if get_char () = 'P' then
                let c = get_char () in
                if c = '5' then
                    (* raw PGM *)
                    return 0 read_byte
                else if c = '2' then
                    (* plain PGM *)
                    return 0 read_dec
                else if c = '4' then
                    (* raw PBM *)
                    return 1 read_bit
                else if c = '1' then
                    (* plain PBM *)
                    return 1 ( fun x -> 1 - read_dec () )
                else if c = '6' then
                    (* raw PPM *)
                    return 0 ( read_rgb read_byte )
                else if c = '3' then
                    (* plain PPM *)
                    return 0 ( read_rgb read_dec )
                else
                    raise ( Invalid_format "wrong magic number" )
            else
                raise ( Invalid_format "wrong magic number" )
        in
        let loader ( width, height, depth, reader ) =
            Image.init width height depth ( fun x y -> reader () )
        in
        loader ( skip_spaces () ; parse_header ) 
end
