(******************************************************************************
*
* simulatedAnnealing.ml: Simulated Annealing algorithm for figure-ground 
*                        discrimination.
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

open Error
open SymmetricMatrix

module SimulatedAnnealing =
    struct

        type spin = Up | Down
        type state = spin array

        (* eval_spin : evaluates a spin *)
        let eval_spin = function 
            | Up ->
                1.0
            | Down -> 
                -.1.0

        (* sum_col : sums over the columns of the matrix j_matrix*)
        let sum_col lambda c_matrix (sigma : state) i =
            let rec _sum_col j acc =
		        if j = SymmetricMatrix.length c_matrix then
		            acc
		        else
                    _sum_col ( j + 1 ) ( ( ( lambda -. ( SymmetricMatrix.get c_matrix i j ) ) *. 
                        ( 
                            if j = i then 
                                0.0 
                            else
                                eval_spin sigma.(j)
                        ) ) +. acc )
            in
            let _ = assert ( ( SymmetricMatrix.length c_matrix ) > 0 ) in
            _sum_col 0 0.0

        (* sum_col_matrix : sums over the columns of the matrix c_matrix*)
        let sum_col_matrix c_matrix i =
            let rec _sum_col_matrix j acc = 
                if  j = SymmetricMatrix.length c_matrix then
                    acc
                else
                    _sum_col_matrix ( j + 1 ) ( ( SymmetricMatrix.get c_matrix i j ) +. acc ) 
            in
                _sum_col_matrix 0 0.0

        (* accept_transition : accepts or rejects a state transition *)
        let accept_transition temp delta_e = 
            ( delta_e < 0.0 ) || ( ( Random.float 1.0 ) < ( exp ( -. delta_e /. temp ) ) )

        (* swap *)      
        let swap sigma_i = 
            match sigma_i with
                | Up ->
                    Down
                | Down ->
                    Up

        (* set_sigma : updates the ith elements of the state vector*)
        let set_sigma ( sigma_old : state ) i =
            let _ = sigma_old.(i) <- ( swap sigma_old.(i) ) in (* SIDEFX *)
            sigma_old

	let create_list n occurr =
	    let rec _create_list i =
            if i = n then
                []
            else
                ( i, occurr ) :: ( _create_list (i + 1))
	    in 
	    _create_list 0
	    
	let create_rndlist n occurr =
	    let list = create_list n occurr in
	    let rec _create_rndlist list list' =
            match list with
                | [] -> 
                    list'
                | head :: tail ->
                    let el = List.nth list ( Random.int ( List.length list ) ) in
                    let ( new_list, _ ) = List.partition ( fun el' -> not ( el' = el ) ) list in
                    _create_rndlist new_list ( el :: list' )
	    in
	    _create_rndlist list []
	    
	let rec get_rndel list =
	    let rec randomize_list list list' =
            match list with
                | [] -> 
                    list'
                | head :: tail ->
                    let el = List.nth list ( Random.int ( List.length list ) ) in
                    let ( new_list, _ ) = List.partition ( fun el' -> not ( el' = el ) ) list in
                    randomize_list new_list ( el :: list' )
	    in
	    match list with
            | [] -> 
                ( -1, [] )
            | ( ind, 0 ) :: tail ->
                get_rndel tail
            | ( ind, num ) :: tail ->
                let new_list = randomize_list ( ( ind, num - 1 ) :: tail ) [] in
                ( ind, new_list )
	    

        (* energy_transition : computes the energy variation associated with an elementary transition*)
        let energy_transition lambda c_matrix sigma i = 
            let sum_lambda = sum_col lambda c_matrix  sigma i in
            let sum_row = sum_col_matrix c_matrix i 
            in
                -. 2.0 *. 
                ( eval_spin sigma.(i) ) *.
                ( sum_lambda +. 
                ( ( float_of_int ( SymmetricMatrix.length c_matrix ) ) *. lambda ) -.
                sum_row )
        
        (* sigma_init : create the state vector*)
        let sigma_init length = 
            let sigma = Array.init length ( fun i -> if ( Random.float 1.0 < 0.5 ) then Down else Up ) 
            in
            sigma

    end

    open SimulatedAnnealing

    (*simulatedAnnealing*)
    let simulatedAnnealing lambda temp num_occ dec_temp sigma c_matrix =
	    let rnd_list = create_rndlist ( Array.length sigma ) num_occ in  
        let rec _simulatedAnnealing temp transition acc_transition iter rnd_list sigma  = 
            let _ = Random.init ( int_of_float ( Unix.time () ) ) in
            let percent = ( float_of_int acc_transition ) *. 100.0 /. ( float_of_int transition ) in
            if ( temp < 0.001 ) then
                let _ = Printf.printf "Annealing completed ( Temperature = 0) \n%!" in
                ( iter, sigma )
            else
                let temp' = temp *. dec_temp in
                (*make a small change in the state, i.e. a state transition*)
                let ( s_index, rnd_list ) = get_rndel rnd_list in
                if ( rnd_list = [] ) then
                    let _ = Printf.printf "\rPercent: %f %!" percent in
                    if ( percent <= 1.0 ) then 
                        (* base of induction *)
                        let _ = Printf.printf "Annealing completed\n%!" in
                        ( iter, sigma )
                    else 
                        _simulatedAnnealing temp' 0 0 ( iter + 1 ) ( create_rndlist ( Array.length sigma ) num_occ ) sigma 
                else
                    let sigma = set_sigma sigma s_index in
                    (* evaluate the resulting change il energy *)
                    let delta_e = energy_transition lambda c_matrix sigma s_index in 
                    let ( acc_transition, sigma ) =    (* accept or reject the transition *)
                        if ( accept_transition temp delta_e ) then
                            ( acc_transition + 1, sigma )
                        else
                            ( acc_transition, ( set_sigma sigma s_index ) )
                    in
                    _simulatedAnnealing temp ( transition + 1 ) acc_transition ( iter + 1 ) rnd_list sigma             
        in
        _simulatedAnnealing temp 0 0 1 rnd_list sigma 
  

 (* get_figure : selects figure edgels  *)
 let get_figure_sim spin_vector edgels =
    let rec _get_figure i edgels figure =
        if ( i = Array.length spin_vector ) then 
            figure
        else
            match ( spin_vector.(i), edgels) with
                | ( Down, head :: tail ) ->
                        _get_figure ( i + 1 ) tail ( head :: figure)
                | ( _, head :: tail ) ->
                        _get_figure ( i + 1 ) tail  figure
                | ( _, _ ) ->
                    let _ = sassert false "get_figure: spin_vector and edgels sizes differ" in []
    in
    _get_figure 0 edgels []
        
    let simulated_annealing lambda c_matrix =
        let _ = Random.init ( int_of_float ( Unix.time () ) ) in
        let sigma = sigma_init ( SymmetricMatrix.length c_matrix ) in
        let ( iter, sigma ) = simulatedAnnealing lambda 100.0 100 0.93 sigma c_matrix in
        ( iter, sigma )       