(******************************************************************************
*
* symmetricMatrix.ml: types and operations for symmetric matrixes
*
* Giulio Manzonetto, Giulio Marcon, Anna Sperotto
* <{gmanzone,gmarcon,asperott}@dsi.unive.it>
*
* 
*******************************************************************************)

module SymmetricMatrix =
   struct
        type 'a sym_matrix = 'a array array

        let length : ('a sym_matrix -> int) = Array.length

        let get (sym_matr : 'a sym_matrix) i j = 
            if i >= j
                then sym_matr.(i).(j)
                else sym_matr.(j).(i)
    
        (* set: sets an element of a matrix and return the modified matrix *)
        let set (sym_matr : 'a sym_matrix) i j value =
            let _ =
                if i >= j
                    then sym_matr.(i).(j) <- value (* SIDEFX *)
                    else sym_matr.(j).(i) <- value (* SIDEFX *)
            in
                sym_matr

        (* updates: updates an element of a symmetric matrix *)
        let update sym_matr f i j = 
            set sym_matr i j ( f ( get sym_matr i j ) )     (* SIDEFX *)
            
        (* init_matrix: creates a new symmetric matrix n x n intialized with the value init_val *)
        let make_matrix n init_val : 'a sym_matrix =
            Array.init n ( fun i -> 
                Array.init ( i + 1 ) ( fun j -> 
                    init_val 
                ) 
            )

        (* init_matrix: creates a new symmetric matrix n x n where a[i,j] is intialized with the function init_fun(i,j) *)
        let init_matrix n init_fun : 'a sym_matrix =
            Array.init n (fun i -> 
                Array.init ( i + 1 ) (fun j -> 
                    init_fun i j
                )
            )

        (* Note: use iteri for matrix upgrade only. iteri returns a FRESH array *)
        let iteri = Array.iteri

        (* Note: use iter for matrix upgrade only iteri returns a FRESH array *)
        let iter = Array.iter

        (* Note: use map for matrix upgrade only *)
        let map (f : 'a -> 'a) a =
            let n = Array.length a in
            let rec _map i =
                if ( i = 0 ) then
                    a.(0) <- f a.(0)
                else
                    (a.(i) <- f a.(i);
                    _map (i - 1))
            in
            _map ( n - 1 );
            a

        (* Note: use map for matrix upgrade only *)
        let mapi (f : int -> 'a -> 'a) a =
            let n = Array.length a in
            let rec _mapi i =
                if ( i = 0 ) then
                    a.(0) <- f 0 a.(0)
                else 
                    (a.(i) <- f i a.(i);
                     _mapi (i - 1))
            in
            _mapi ( n - 1 );
            a

        (* fold_left_matrix: executes a fold_left on a (n x n)-sym_matrix *)
        let fold_left_matrix f x (matrix : 'a sym_matrix) =
            let n = Array.length matrix in
            let rec _fold_left_matrix f i j =
                if ( ( i = 0 ) && ( j = 0 ) ) then
                    f x ( get matrix 0 0 )
                else
                    if ( j >= 0 ) then
                        f (_fold_left_matrix f i ( j - 1 ) ) ( get matrix i j )
                    else
                        _fold_left_matrix f  ( i - 1 ) ( n - 1 )
            in
                _fold_left_matrix f ( n - 1 ) ( n - 1 )


        (* print_matrix: prints a matrix as it is saved in memory *)
        let print_matrix (matrix : 'a sym_matrix ) =
            Array.iter (fun row -> 
                let _ = 
                    Array.iter (fun el -> 
                        Printf.printf " %f " el
                    ) row
                in
                    Printf.printf "\n"   (* SIDEFX *)
            ) matrix

        (* print_sym_matrix: prints a symmetric matrix *)
        let print_sym_matrix (matrix : 'a sym_matrix ) =
            for i = 0 to (Array.length matrix) - 1 do
                for j = 0 to (Array.length matrix) - 1 do
                    Printf.printf " %f" ( get matrix i j )    (* SIDEFX *)
                done;
                Printf.printf "\n";
            done;
   end

module SymmetricMatrix_without_sidefx =
   struct
        type 'a sym_matrix = 'a array array

        let length : ('a sym_matrix -> int) = Array.length

        let get (sym_matr : 'a sym_matrix) i j = 
            if i >= j
                then sym_matr.(i).(j)
                else sym_matr.(j).(i)

        (* set: sets an element of a matrix and return a FRESH modified matrix *)
        let set (sym_matr : 'a sym_matrix) i j value =
            let sym_matr' = Array.copy sym_matr in
            ( if i >= j
                then sym_matr'.(i).(j) <- value    (* SIDEFX *)
                else sym_matr'.(j).(i) <- value    (* SIDEFX *)
              ; sym_matr' )

        (* updates: updates an element of a symmetric matrix *)
        let update sym_matr f i j =
            let sym_matr' = Array.copy sym_matr
            in
                set sym_matr i j (f (get sym_matr i j))

        (* init_matrix: creates a new symmetric matrix n x n intialized with the value init_val *)
        let make_matrix n init_val : 'a sym_matrix =
            Array.init n (fun i ->
                Array.init ( i + 1 ) (fun j ->
                    init_val
                )
            )

        (* init_matrix: creates a new symmetric matrix n x n where a[i,j] is intialized with the function init_fun(i,j) *)
        let init_matrix n init_fun : 'a sym_matrix =
            Array.init n (fun i ->
                Array.init ( i + 1 ) (fun j ->
                    init_fun i j
                )
            )

        (* Note: use iteri for matrix upgrade only *)
        let iteri = Array.iteri

        (* Note: use iter for matrix upgrade only *)
        let iter = Array.iter

        (* Note: use map for matrix upgrade only *)
        let map : (('a -> 'a) -> 'a array -> 'a array) = Array.map

        (* Note: use map for matrix upgrade only *)
        let mapi : ((int -> 'a -> 'a) -> 'a array -> 'a array) = Array.mapi

        (* fold_left_matrix: executes a fold_left on a (n x n)-sym_matrix *)
        let fold_left_matrix f x (matrix : 'a sym_matrix) =
            let n = Array.length matrix in
            let rec _fold_left_matrix f i j =
                if ( ( i = 0 ) && ( j = 0 ) ) then
                    f x (get matrix 0 0)
                else
                    if ( j >= 0 ) then
                        f ( _fold_left_matrix f i ( j - 1 ) ) ( get matrix i j )
                    else
                        _fold_left_matrix f  ( i - 1 ) ( n - 1 )
            in
                _fold_left_matrix f ( n - 1 ) ( n - 1 )


        (* print_matrix: prints a matrix as it is saved in memory *)
        let print_matrix (matrix : 'a sym_matrix ) =
            Array.iter (fun row -> 
                let _ = 
                    Array.iter (fun el -> 
                        Printf.printf " %f " el
                    ) row 
                in
                    Printf.printf "\n"   (* SIDEFX *)
            ) matrix

        (* print_sym_matrix: prints a symmetric matrix *)
        let print_sym_matrix (matrix : 'a sym_matrix ) =
            for i = 0 to ( Array.length matrix ) - 1 do
                for j = 0 to (Array.length matrix) - 1 do
                    Printf.printf " %f" (get matrix i j)    (* SIDEFX *)
                done;
                Printf.printf "\n";
           done;
   end

